/* SACHICA: Scalable Algorithm for Characteristic and Homologinous Interval CAlclation */
/* 2004 Takeaki Uno */
/* filtering isolated points from the file */

#ifndef _sachicax_c_
#define _sachicax_c_

#define STDLIB2_RADIX_SORT    // use fast integer radix sort, written in stdlib2.c

#include"stdlib2.c"
#include"pos.c"
#include"problem.c"

#define SACHICAX_REPRESENTATIVE 65536  // output representative of each similar string of the length
#define SACHICAX_IGNORE_SHORT 131072  // ignore pairs concentrated in a small area
#define SACHICAX_DISTANCE_FILTER 262144  // execute distance filtering

/*************************/
void SACHICAx_error (){
  ERROR_MES = "command explanation";
  print_err (" sachica-extract: extract the similar substrings in the area of long similar substringsoutput by sachica\n\
sachicax [option] input-filename length number width output-filename\n\
[options]\n\
-X: output representative of pairs in the same area\n\
-Z: ignore pairs in a small box\n\
-r [num]: specify the length of the small box (for -Z, default is [width])\n\
-D: switch the filtering method to distance filtering, with factor [length]\n\
-n [num]: give the number of pair\n\
# the 1st letter of input-filename cannot be '-'.\n");
  EXIT;
}

/************************************************************************/
void SACHICAx_read_param (PROBLEM *PP, int argc, char *argv[]){
  int c=1;
  PP->dim = 2;    // integers for one object
  PP->gap_ub = 0;

  if ( argc < c+3 ){ SACHICAx_error (); return; }

  while ( strchr ( argv[c], '-') ){
    switch (argv[c][1]){
      case 'X': c--; PP->problem |= SACHICAX_REPRESENTATIVE;
      break; case 'D': c--; PP->problem |= SACHICAX_DISTANCE_FILTER; PP->dim = 3;
      break; case 'Z': c--; PP->problem |= SACHICAX_IGNORE_SHORT;
      break; case 'n': if ( (PP->II.outputs = atoi(argv[c+1])) < 0 )
          error_num("#pairs has to be non-negative", atoi(argv[c+1]), EXIT);
      break; case 'r': if ( (PP->gap_ub = atoi(argv[c+1])) < 0 )
          error_num("#pairs has to be non-negative", atoi(argv[c+1]), EXIT);
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+3 ){ SACHICAx_error (); return; }
  }

  NEXT:;
  PP->input_fname = argv[c];
  if ( PP->problem & SACHICAX_DISTANCE_FILTER ){
    if ( (PP->ratio = atof(argv[c+1])) < 0 )
    error_num("distance factor for distance filtering has to be non-negative", PP->ratio, EXIT);
  } else {
    if ( argc < c+5 ){ SACHICAx_error (); return; }
    if ( (PP->len = atoi(argv[c+1])) <= 0 )   // length
       error_num("box filter length has to be positive",atoi(argv[c+1]), EXIT);
    PP->perr = atoi(argv[c+2]);     // threshold
    if ( (PP->width = atoi(argv[c+3])) <= 0 )  // width
       error_num("box filter width has to be positive",atoi(argv[c+3]), EXIT);
    c += 2;
  }
  PP->output_fname = argv[c+2];
  if ( PP->gap_ub == 0 ) PP->gap_ub = PP->width;
  print_err ("SACHICAX: input_file= %s ,output_file= %s ,len= %d ,width= %d ,th= %d, small_box= %d\n", PP->input_fname, PP->output_fname, PP->len, PP->width, PP->perr, PP->gap_ub);
}

/* read from file */
void SACHICAx_read_file (PROBLEM *PP){
  FILE2 fp;
  SEQ_LONG x, y, *v;
  size_t i, k=0;
  
  PP->xmax = PP->ymax = 0;
    // count #pairs
  if ( PP->input_fname ){
    FILE2_open (fp, PP->input_fname, "r", EXIT);
  } else fp.fp = stdin;

  if ( PP->II.outputs > 0 ) k = (size_t)PP->II.outputs; 
  else {
    while (1){   // count #out-going edges for each vertex
      x = (SEQ_LONG)FILE2_read_int (&fp);
      if ( FILE_err&2 ) break;
      if (FILE_err&4) continue;
      y = (SEQ_LONG)FILE2_read_int (&fp);
      if (FILE_err&4) continue;
      k++;
      FILE2_read_until_newline (&fp);
    }
    print_err ("sachicax: scan end\n");  
    
    if ( PP->input_fname ) FILE2_reset (&fp);
  }
  
  // malloc for edges and edge-lists
  malloc2 (PP->SS.head, (k+3)*PP->dim, {FILE2_close(&fp);EXIT;}); v=PP->SS.head;

    // read pairs from file
  FLOOP (i, 0, k){
    x = (SEQ_LONG)FILE2_read_int (&fp);
    if ( FILE_err&2 ) break;
    if (FILE_err&4) continue;
    y = (SEQ_LONG)FILE2_read_int (&fp);
    if (FILE_err&4) continue;
    if ( PP->dim > 2 ) v[i*PP->dim+2] = (SEQ_LONG)FILE2_read_int (&fp);
    FILE2_read_until_newline (&fp);
    ENMAX (PP->ymax, y+1); ENMAX (PP->xmax, x+1);
    v[i*PP->dim] = x;
    v[i*PP->dim+1] = y;
  }
  FILE2_close (&fp);
  if (PP->ymax == 0) return;
  
print_err ("sachicax: read end, start sorting...\n");  
  FLOOP (i, 0, k) v[i*PP->dim] += PP->ymax - v[i*PP->dim+1];
// sort. need to be modified if SEQ_LONG != int
//  intarray_sort (v, k, sizeof(int)*2);
  qsort_SEQ_LONG (v, k, sizeof(int)*PP->dim);
  FLOOP (i, 0, k) v[i*PP->dim] -= PP->ymax - v[i*PP->dim+1];
  PP->num = k;    // memorize #tuples

print_err ("sachicax: sort end\n");  
}

/* adjast xmax and ymax */
void SACHICAx_adj_max (PROBLEM *PP){
  size_t i;
  PP->pymax = INTHUGE;
  FLOOP (i, 0, PP->num){
    ENMAX (PP->xmax, PP->SS.head[i*3]+PP->SS.head[i*3+2]+1);
    ENMAX (PP->ymax, PP->SS.head[i*3+1]+PP->SS.head[i*3+2]+1);
    ENMAX (PP->pxmax, PP->SS.head[i*3+2]);
    ENMIN (PP->pymax, PP->SS.head[i*3+2]);
  }
}


/* get the value of "x+y" with considering the flag, added to the small mximum column */
SEQ_LONG SACHICAx_xy (PROBLEM *PP, size_t i){
  i *= PP->dim;
  return (PP->SS.head[i] + PP->SS.head[i+1]);
}
SEQ_LONG SACHICAx_x_y (PROBLEM *PP, size_t i){
  i *= PP->dim;
  return (PP->SS.head[i] + PP->ymax - PP->SS.head[i+1]);
}

/* sort segments in PP->SS.head, where segment is composed of points in a diagonal line of width w */
void SACHICAx_sort_stabs (PROBLEM *PP, int w){
  size_t i, j=0;
  SEQ_LONG base = ((int)(SACHICAx_x_y (PP, 0)/w)) *w;

  PP->SS.head[0] += PP->SS.head[1];
  FLOOP (i, 1, PP->num+1){
    if ( i<PP->num && SACHICAx_x_y (PP, i) < base + w ) goto LOOP_END;
    qsort_SEQ_LONG (PP->SS.head+j*PP->dim, i-j, sizeof(SEQ_LONG)*PP->dim);
    j = i; base = ((int)(SACHICAx_x_y (PP, i)/w)) *w;
    LOOP_END:;
    PP->SS.head[i*PP->dim] += PP->SS.head[i*PP->dim+1];
  }
  FLOOP (i, 0, PP->num) PP->SS.head[i*PP->dim] -= PP->SS.head[i*PP->dim+1];
}

/* write a tuple of position p to the file fp */
void SACHICAx_write_tuple (PROBLEM *PP, FILE2 *fp, size_t p){
  SEQ_LONG *n = &PP->SS.head[p*PP->dim];
  FILE2_print_int (fp, *n, 0); n++;
  FILE2_print_int (fp, *n, ' '); n++;
  if ( PP->dim == 3 ) FILE2_print_int (fp, *n, ' ');
  FILE2_putc (fp, '\n');
  FILE2_flush (fp);
}

/* filter isolated pairs */
void SACHICAx_filter (PROBLEM *PP){
  size_t p, pp, s1, s2, s3, i1, i2, j1, j2, h1, h2;
  SEQ_LONG x, y, z;
  SEQ_LONG base = ((int)(SACHICAx_x_y(PP, 0)/PP->width)) *PP->width;

  for (s1=s2=s3=0 ; s1 < PP->num ; base+=PP->width){
    for ( ; s3<PP->num && SACHICAx_x_y (PP, s3) < base +PP->width ; s3++);
    for (i1=j1=h1=s1,i2=j2=h2=s2,y=0 ; i1<PP->num && SACHICAx_x_y(PP,i1)<base+PP->width ; ){
        // find the minimum
      pp = i1; x = SACHICAx_xy (PP, i1);
      if ( i2 < s1 && (z=SACHICAx_xy (PP, i2)) < x ){ pp = i2; x = z; }
      
        // find the farthest within the length, for both stab
      while ( j2 < s1 && (z=SACHICAx_xy (PP, j2)) < x +PP->len ){ ENMAX (y, z); j2++; }
      while ( j1 < s3 && (z=SACHICAx_xy (PP, j1)) < x +PP->len ){ ENMAX (y, z); j1++; }
      if ( (j1-i1) + (j2-i2) >= (size_t)PP->perr &&
         (!(PP->problem&SACHICAX_IGNORE_SHORT) || y - x > (SEQ_LONG)PP->gap_ub)){
        if ( PP->problem & SACHICAX_REPRESENTATIVE ){
            // the minimum will be the representative, and is output
          SACHICAx_write_tuple (PP, &PP->II.multi_fp[0], pp);
          i1 = j1; i2 = j2; // skip all the intermediate points
        } else {
          FLOOP (p, MAX(i2,h2), j1){
            if ( p >= j2 && p<i1 ){ p = MAX(i1,h1); if ( p>=j1 ) break; } // jump to stab 1
            BIT_SET (PP->vf, p);
          }
          h1 = j1; h2 = j2;
        }
      }
      if ( pp == i2 ) i2++; else i1++;   // increment the counter of the minimum
    }
    s2 = s1; s1 = i1;  // set to the head of the next slots
  }

  if ( !(PP->problem & SACHICAX_REPRESENTATIVE) ){
    FLOOP (p, 0, PP->num)
      if ( ISBIT (PP->vf, p) ) SACHICAx_write_tuple (PP, &PP->II.multi_fp[0], p);
  }
}

/* filter isolated pairs */
void SACHICAx_distance_filter (PROBLEM *PP, int l, int u){
  size_t p, s1, s3, i1, i2, j;
  SEQ_LONG x, z, a=PP->dim, xx, yy, *v = PP->SS.head;
  SEQ_LONG base = ((int)(SACHICAx_x_y (PP, 0)/PP->width)) *PP->width;

  SACHICAx_sort_stabs (PP, PP->width);
  for (s1=i1=i2=s3=0 ; s1 < PP->num ; base+=PP->width){
    for ( ; s3<PP->num && SACHICAx_x_y (PP, s3) < base +PP->width ; s3++);
    while (i1 < s3){

        // find the minimum and increment the one with the minimum
      p = i1; x = SACHICAx_xy (PP, i1);
      if ( i2 < s1 && (z=SACHICAx_xy (PP, i2)) < x ){ p = i2; x = z; i2++; } else i1++;

        // segment length is not to be treated
      z = v[p*a+2] & 0x3fffffff;
      if ( v[p*a+2] & 0x40000000 || !RANGE (l+1, z, u) ) continue;

      x += z*2 -2 +z*PP->ratio;  // set x to the tail of the segment
      xx = v[p*a] + z -1;  // set xx to the x-coordinate of tail
      yy = v[p*a+1] + z -1;  // set xx to the y-coordinate of tail
      
        // find the farthest within the length, for both stab
      if ( SACHICAx_x_y(PP, p) >= base ) for (j=i2*a ; j<s1*a ; j+=a){
        if ( v[j]+v[j+1] > x ) break;
        if ( v[j] >= xx && v[j+1] >= yy ){ v[p*a+2] |= 0x40000000; goto END;}
      }
      for (j=i1*a ; j<s3*a ; j+=a){
        if ( v[j]+v[j+1] > x ) break;
        if ( v[j] >= xx && v[j+1] >= yy ){ v[p*a+2] |= 0x40000000; goto END;}
      }
      END:;
    }
    i2 = s1; s1 = i1;  // set to the head of the next slots
  }

    // change all segments to the tail position
  FLOOP (p, 0, PP->num){
    v[p*a] += (v[p*a+2] & 0x3fffffff) -1;
    v[p*a+1] += (v[p*a+2] & 0x3fffffff) -1;
  }

  SACHICAx_sort_stabs (PP, PP->width);
  base = ((int)(SACHICAx_x_y (PP, PP->num-1)/PP->width)) *PP->width;
  for (s1=i1=i2=s3=PP->num ; s1>0 ; base-=PP->width){
    for (s3=s1 ; s3>0 && SACHICAx_x_y(PP,s3-1)>=base ; s3--);
    while (i1 > s3){
        // find the minimum and increment the one with the minimum
      p = i1-1; x = SACHICAx_xy (PP, p);
      if ( i2 > s1 && (z=SACHICAx_xy (PP, i2-1)) > x ){ p = i2-1; x = z; i2--; } else i1--;
        // segment length is not to be treated
      z = v[p*a+2] & 0x3fffffff;
      if ( v[p*a+2] & 0x40000000 || !RANGE (l+1, z, u) ) continue;

      x -= z*2 -2 +z*PP->ratio;  // set x to the head of the segment
      xx = v[p*a] - (z -1);  // set xx to the x-coordinate of head
      yy = v[p*a+1] - (z -1);  // set xx to the y-coordinate of head
        // find the farthest within the length, for both stab
      if ( SACHICAx_x_y(PP, p) >= base ) for (j=i2*a ; j>s1*a ; ){ j-=a;
        if ( v[j]+v[j+1] < x ) break;
        if ( v[j] <= xx && v[j+1] <= yy ){ v[p*a+2] |= 0x40000000; goto END2;}
      }
      for (j=i1*a ; j>s3*a ; ){ j-=a;
        if ( v[j]+v[j+1] < x ) break;
        if ( v[j] <= xx && v[j+1] <= yy ){ v[p*a+2] |= 0x40000000; goto END2;}
      }
      END2:;
    }
    i2 = s1; s1 = i1;  // set to the head of the next slots
  }

    // change all segments to the head position
  FLOOP (p, 0, PP->num){
    v[p*a] -= (v[p*a+2] & 0x3fffffff) -1;
    v[p*a+1] -= (v[p*a+2] & 0x3fffffff) -1;
  }
}


/* main routine */
int SACHICAx_main (int argc, char *argv[]){
  PROBLEM PP;
  size_t p;

  PROBLEM_init (&PP);
  SACHICAx_read_param (&PP, argc, argv);
if ( ERROR_MES ) goto END;
  SACHICAx_read_file (&PP);
if ( ERROR_MES ) goto END;
  if ( PP.num == 0 ) goto END;
  malloc2 (PP.II.multi_fp, 1, EXIT);
  if ( !strcmp (PP.output_fname,"-") ) PP.II.multi_fp[0].fp = stdout;
  else FILE2_open (PP.II.multi_fp[0], PP.output_fname, "w", EXIT);

  if ( PP.problem & SACHICAX_DISTANCE_FILTER ){
    SACHICAx_adj_max (&PP);
    PP.width = PP.pymax*2 *PP.ratio;
    SACHICAx_distance_filter (&PP, 0, PP.pymax*2);
    PP.width *= (int)((PP.pxmax+PP.pymax*2-1)/(PP.pymax*2));
    SACHICAx_distance_filter (&PP, PP.pymax*2, PP.pxmax+1);
    FLOOP (p, 0, PP.num){
      if ( PP.SS.head[p*3+2] & 0x40000000 ){
        PP.SS.head[p*3+2] -= 0x40000000;
        SACHICAx_write_tuple (&PP, &PP.II.multi_fp[0], p);
      }
    }
  } else {
    calloc2 (PP.vf, PP.num/32+1, EXIT);
    SACHICAx_sort_stabs (&PP, PP.width);
    SACHICAx_filter (&PP);
  }

  FILE2_closew (&PP.II.multi_fp[0]);

  END:;
  free2 (PP.SS.head);
  PROBLEM_end (&PP);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (SACHICAx_main (argc, argv) );
}
#endif
/*******************************************************************************/


#endif

