/************************************************************************/
/* sheap.c: Similarity/Homology Enhanced Analyze Procedure              */
/* analyze the similarity structure of the given string and visualize it 
   as a picture (dot-plot, and frequency analysis) */
/* 20/Dec/2009 Takeaki Uno  */

#ifndef _sheap_c_
#define _sheap_c_

#ifdef _NO_MAIN_
#define _sheap_c_no_main_
#else
#define _NO_MAIN_   // for removing "main routine" from each file
#endif

#define STDLIB2_RADIX_SORT    // use fast integer radix sort, written in stdlib2.c

#include<string.h>
#include"sachica.c"
#include"sachicam.c"
#include"sachicax.c"
#include"unionpgm.c"


int SHEAP_innum=3, SHEAP_box_factor=100, SHEAP_horizon=1000, SHEAP_emphasis=0;
int SHEAP_flag=0, SHEAP_len=0, SHEAP_bmp=0, SHEAP_plen, SHEAP_mis, SHEAP_line_skip=4;
int SHEAP_repeat=0, SHEAP_negative=0, SHEAP_draw_th=0, SHEAP_sachicax=1, SHEAP_cell_siz=0;
int SHEAP_genome=0, SHEAP_mes=SHOW_MESSAGE;
double SHEAP_bright=4.0, SHEAP_smallbox_factor=0.0, SHEAP_dist_factor=0.0;
char *SHEAP_infname, *SHEAP_outfname, *SHEAP_workdir, SHEAP_workdir_tmp[2] = {0,0};
int SHEAP_no_remove=0;

/* error routine */
void SHEAP_error (){
  ERROR_MES = "command explanation";
  print_err ("sheap: tGiBNRX [options] input-file substring-length #mismatch output-file\n\
%%:show progress, _:no message, =:do not delete temporal files\n\
t:ascii text comparison mode, G:genome sequence comparison mode\n\
i:ignore newline code, B:output in bmp (bitmap) format\n\
N:negative image, R:repeat string detection, X:do not execute box-filtering\n\
x:put a representative for each box, for box-filtering\n\
\n[options]\n\
-2 [filename]:compare 2 files (input-file and [filename])\n\
-h [num]:set horizontal pixel size to [num]\n\
-p [num]:set the cell size to [num]\n\
-d [real-num]:set brightness to [real-num]\n\
-l [num]:specify gap between two horizontal lines\n\
-I [num]:specify interleave length to [num]\n\
-S [char]:ignore substrings including [#mismatch] of [char]\n\
-Z [num]:emphasize each dot to size of [num]\n\
-t [num]:change the drawing mode to threshold drawing and specify threshold to [num] (put a dot when [num] number of pairs are in a pixel)\n\
-T [num]:specify threshold num for filtering (pick up the similar pairs only when at least this number of similar pairs are in a box of the box-filtering)\n\
-X [num]:specify the factor of box for filtering (default is 100, so that the length/width of the box will be substring-length*100/substring-length*10)\n\
-r [num]:set the size of the small box to [num]*[length of the box] (default is [width])\n\
-D [num]:use distance filtering, with factor [num]\n\
-H [num]:extract homolog of length [len]\n\
-M [num]:use [num] cores of CPU\n\
-W [dir]:specify the working directory (folder). The last letter of the directory has to be '/' ('\\')\n\
# the 1st letter of input-filename cannot be '-'.\n");
  EXIT;
}

/* read parameters given by command line */
void SHEAP_read_param (int argc, char *argv[], char *comm){
  int c=1;
  char *z;
  SHEAP_workdir = SHEAP_workdir_tmp;

  if ( argc<c+5 ){ SHEAP_error (); return; }
  comm[0] = 0;
  strcat (comm, "ppppppll");
  if ( strchr( argv[1], '_' ) ){ strcat (comm, "_"); SHEAP_mes = 0; }
  if ( strchr( argv[1], '%' ) ) strcat (comm, "%");
  if ( strchr( argv[1], '=' ) ) SHEAP_no_remove = 1;
  if ( strchr( argv[1], 'i' ) ) strcat (comm, "i");
  if ( strchr( argv[1], 'B' ) ) SHEAP_bmp = 1;
  if ( strchr( argv[1], 'N' ) ) SHEAP_negative = 1;
  if ( strchr( argv[1], 'R' ) ){ comm[0] = comm[1] = comm[2] = 'r' ; SHEAP_repeat = 1; }
  if ( strchr( argv[1], 'X' ) ) SHEAP_sachicax = 0;
  if ( strchr( argv[1], 'x' ) ) SHEAP_sachicax = 2;
  c++;

  while ( argv[c][0] == '-' ){
    if ( argc<c+5 ){ SHEAP_error (); return; }
    switch ( argv[c][1] ){
      case 'T': if ( (SHEAP_innum = atoi(argv[c+1])) <= 0 )
          error_num("threashold has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'D': if ( (SHEAP_dist_factor = atof(argv[c+1])) <= 0 )
          error_num("factor for distance filtering has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'X': if ( (SHEAP_box_factor = atoi(argv[c+1])) <= 0 )
          error_num("filtering box size has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'r': if ( (SHEAP_smallbox_factor = atof(argv[c+1])) <= 0 )
          error_num("filtering box size has to be positive",atof(argv[c+1]), EXIT);
      break; case 't': if ( (SHEAP_draw_th = atoi(argv[c+1])) <= 0 )
          error_num("draw threashold has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'h': if ( (SHEAP_horizon = atoi(argv[c+1])) <= 0 )
          error_num("horizontal pixel size has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'p': if ( (SHEAP_cell_siz = atoi(argv[c+1])) <= 0 )
          error_num("cell size has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'Z': if ( (SHEAP_emphasis = atoi(argv[c+1])) < 0 )
          error_num("emphasis factor has to be a non-negative number",atoi(argv[c+1]), EXIT);
      break; case 'H': if ( (SHEAP_len = atoi(argv[c+1])) <= 0 )
          error_num("the threshold for homolog length has to be positive", atoi(argv[c+1]), EXIT);
      break; case 'd': if ( (SHEAP_bright = atof(argv[c+1])) <=0 )
          error_num("threashold has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'l': if ( (SHEAP_line_skip = atoi(argv[c+1])) <= 0 )
          error_num("interval length of lines has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'I': strcat (comm, " -I "); strcat (comm, argv[c+1]);
      break; case 'S': strcat (comm, " -S "); strcat (comm, argv[c+1]);
      break; case 'M': strcat (comm, " -M "); strcat (comm, argv[c+1]);
      break; case 'W': SHEAP_workdir = argv[c+1];
      break; case '2': strcat (comm, " -n -1 -c -1 -2 ");
          strcat (comm, argv[c+1]); strcat (comm, " ");
        SHEAP_flag = 1; comm[0] = 'c';
      break; default: for (z="SqQxbB" ; *z!=0 ; z++){
          if ( strchr(argv[c], *z) ){
            strcat (comm, " "); strcat (comm, argv[c]);
            strcat (comm, " "); strcat (comm, argv[c+1]);
          }
        }
    }
    c += 2;
  }

  if ( SHEAP_dist_factor > 0.0 ) comm[7] = 'Z';  // maximal substrings for dist-filt.
  if ( strchr( argv[1], 'G' ) ){
    comm[1] = 'G'; strcat (comm, " -S N");
    if ( SHEAP_repeat>0 ) strcat (comm, SHEAP_flag? " -d 1 -r -2 -o -1": " -d 1 -r -2 -o -2");
    SHEAP_genome = 1;
  }
  SHEAP_infname = argv[c];    // input file name
  SHEAP_plen = atoi(argv[c+1]);  // length
  SHEAP_mis = atoi(argv[c+2]);  // #mismatches
  SHEAP_outfname = argv[c+3];  // output file name
  if ( SHEAP_cell_siz > 0 ) SHEAP_horizon = SHEAP_cell_siz;
}

/******************************************************************************/
int SHEAP_comp_main (char *sachica_comm){
  int sep_by, chunk, width, flag=1, G=SHEAP_genome;
  char s1[20], s2[20], *W=SHEAP_workdir;
  
///////////////////////////////////////////////////////////

//  EXECSUB (SACHICA_main, goto END, "sachica %s -I %d -k 250 -h \"%s__SHEAP_pos__\" \"%s\" %d %d \"%s__SHEAP_out__\"",
  EXECSUB (SACHICA_main, SHEAP_mes, goto END, "sachica %s -h \"%s__SHEAP_pos__\" \"%s\" %d %d \"%s__SHEAP_out__\"",
      sachica_comm, SHEAP_workdir, SHEAP_infname, SHEAP_plen, SHEAP_mis, SHEAP_workdir);
  sep_by = internal_params.i6;

///////////////////////////////////////////////////////////
  chunk = SHEAP_plen*SHEAP_box_factor;
  width = SHEAP_plen*SHEAP_box_factor/10;
  if ( SHEAP_sachicax ){
    sprintf (s2, " -r %d", (int)(chunk*SHEAP_smallbox_factor));
    if ( SHEAP_dist_factor > 0.0 ){
      EXECSUB (SACHICAx_main, SHEAP_mes, goto END, "sachicax -D -n " LONGF " \"%s__SHEAP_out__\" %f \"%s__SHEAP_outx__\"",
        internal_params.l2, W, SHEAP_dist_factor, W);
    } else {
      EXECSUB (SACHICAx_main, SHEAP_mes, goto END, "sachicax -Z%s%s -n " LONGF " \"%s__SHEAP_out__\" %d %d %d \"%s__SHEAP_outx__\"",
        SHEAP_sachicax==2?" -X":"", SHEAP_smallbox_factor>0.0?s2:"", internal_params.l2, W, chunk, SHEAP_innum, width, W);
    }
    REMOV (W, "__SHEAP_out__");
    RENAM (W, "__SHEAP_outx__", "__SHEAP_out__");
  }
///////////////////////////////////////////////////////////
  if ( G ){
//    EXECSUB (SACHICA_main, goto END, "sachica c%s -I %d -n -1 -c -1 -d %d -r -1 -k 250 \"%s\" %d %d \"%s__SHEAP_outr__\"",
    EXECSUB (SACHICA_main, SHEAP_mes, goto END, "sachica c%s -n -1 -c -1 -d %d -r -1 \"%s\" %d %d \"%s__SHEAP_outr__\"",
        sachica_comm, SHEAP_flag?-1:1, SHEAP_infname, SHEAP_plen, SHEAP_mis, W);
      ///////////////////////////////////////////////////////////
    if ( SHEAP_sachicax ){
      if ( SHEAP_dist_factor > 0.0 ){
        EXECSUB (SACHICAx_main, SHEAP_mes, goto END, "sachicax -D -n " LONGF " \"%s__SHEAP_outr__\" %f \"%s__SHEAP_outrx__\"",
          internal_params.l2, W, SHEAP_dist_factor, W);
      } else {
        EXECSUB (SACHICAx_main, SHEAP_mes, goto END, "sachicax -Z%s%s -n " LONGF " \"%s__SHEAP_outr__\" %d %d %d \"%s__SHEAP_outrx__\"",
          SHEAP_sachicax==2?" -X":"", SHEAP_smallbox_factor>0.0?s2:"", internal_params.l2, W, chunk, SHEAP_innum, width, W);
      }
      REMOV (W, "__SHEAP_outr__");
      RENAM (W, "__SHEAP_outrx__", "__SHEAP_outr__");
    }
  }

    //   intensity drawing
  s1[0] = 0; if ( SHEAP_flag ) sprintf (s1, "-c %d", sep_by);

  EXECSUB (SACHICAm_main, SHEAP_mes, goto END, "sachicam -%c %f -Z %d%s%s%s %s%s%s%s -H \"%s__SHEAP_pos__\" \"%s__SHEAP_out__\" %d %d \"%s.%s\"",
      SHEAP_draw_th>0?'t':'k', SHEAP_draw_th>0?(double)SHEAP_draw_th:SHEAP_bright, SHEAP_emphasis,
      SHEAP_flag?"":" -M", SHEAP_negative?" -N":"",
      SHEAP_bmp?" -p":"", s1, G?" -R ":"", G?W:"", G?"__SHEAP_outr__":"", W, W, SHEAP_horizon,
      SHEAP_cell_siz>0?-1:(SHEAP_flag?0:SHEAP_horizon), SHEAP_outfname, SHEAP_bmp?"bmp":"pgm");

///////////////////////////////////////////////////////////
  if ( SHEAP_len > 0 ){
      //  extract homology
    EXECSUB (SACHICAm_main, SHEAP_mes, goto END, "sachicam -t 1 -T %d %s -H \"%s__SHEAP_pos__\" -d \"%s__SHEAP_homo__\" \"%s__SHEAP_outx__\" %d %d \"%s__SHEAP_homo__.pgm\"",
       SHEAP_len, s1, W, W, W, SHEAP_horizon, SHEAP_cell_siz>0?-1:(SHEAP_flag?0:SHEAP_horizon), W);

      //   extract homology; reverse direction
    EXECSUB (SACHICAm_main, SHEAP_mes, goto END, "sachicam -t 1 -T %d %s -H \"%s__SHEAP_pos__\" -E \"%s__SHEAP_homor__\" \"%s__SHEAP_outrx__\" %d %d \"%s__SHEAP_homor__.pgm\"",
        SHEAP_len, s1, W, W, W, SHEAP_horizon, SHEAP_flag?0:SHEAP_horizon, W);

      //  superimpose the homology extracted images
    EXECSUB (UNIONPGM_main, SHEAP_mes, goto END, "unionpgm%s -R \"%s__SHEAP_homo__.pgm\" \"%s__SHEAP_homor__.pgm\" \"%sh.%s\"",
         SHEAP_bmp?" -p":"", W, W, SHEAP_outfname, SHEAP_bmp?"bmp":"pgm");
  }
  flag = 0;

///////////////////////////////////////////////////////////
  END:;
  if ( !SHEAP_no_remove ) MREMOV (W, "__SHEAP_out__", "__SHEAP_outr__", "__SHEAP_pos__", "__SHEAP_homo__", "__SHEAP_homor__", "__SHEAP_homo__.pgm", "__SHEAP_homor__.pgm");
  return (flag);
}

/******************************************************************************/
int SHEAP_repeat_main (char *sachica_comm){
  char *W = SHEAP_workdir;

///////////////////////////////////////////////////////////
  sachica_comm[0] = 'r';

  EXECSUB (SACHICA_main, SHEAP_mes, goto END, "sachica ZMv%s -k 1 -l %d -h \"%s__SHEAP_pos__\" \"%s\" %d %d \"%s__SHEAP_out__\"",
      sachica_comm, SHEAP_plen*2, W, SHEAP_infname, SHEAP_plen, SHEAP_mis, W);


    ///////////////////////////////////////////////////////////
  EXECSUB (SACHICAm_main, SHEAP_mes, goto END, "sachicam -%c %f%s -A -l %d%s -H \"%s__SHEAP_pos__\" \"%s__SHEAP_out__\" %d 0 \"%s.%s\"",
      SHEAP_draw_th>0?'t':'k', SHEAP_draw_th>0?(double)SHEAP_draw_th:SHEAP_bright,
      SHEAP_bmp?" -p":"", SHEAP_line_skip,
      SHEAP_negative?" -N":"", W, W, SHEAP_horizon, SHEAP_outfname, SHEAP_bmp?"bmp":"pgm");

///////////////////////////////////////////////////////////
  END:;
  if ( !SHEAP_no_remove ) MREMOV (W, "__SHEAP_out__", "__SHEAP_pos__");
  return (0);
}

/*******************************************************************************/
int SHEAP_main (int argc, char *argv[]){
  char sachica_comm[100];
  SHEAP_read_param (argc, argv, sachica_comm);
  if ( ERROR_MES ) return (1);
  if ( SHEAP_repeat>0 ) return (SHEAP_repeat_main (sachica_comm));
  else return (SHEAP_comp_main (sachica_comm));
}

/*******************************************************************************/
#ifndef _sheap_c_no_main_
int main (int argc, char *argv[]){
  return (SHEAP_main (argc, argv) );
}
#endif
/*******************************************************************************/

#endif




